-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Globals.lua
-- DESC: Addon-scope variables.

local addonName, HoJ = ...

-- HoJ.ADDON_VERSION = 0.5; -- split hopo window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.6; -- split icon window from Templar's Verdict window
-- HoJ.ADDON_VERSION = 0.81; -- major rewrite for Mists of Pandaria with low level pally fixes
-- HoJ.ADDON_VERSION = 0.82; -- ret skill bar for rebuke, blinding light, hammer of justice / fist of justice, added selfless healer to buffs, added stack count to buffs and skills. Did not implement skills for Prot yet.
-- HoJ.ADDON_VERSION = 0.9; -- Warlord's of Draenor. Removed skills that no longer exists (inquisition, blinding light)
-- HoJ.ADDON_VERSION = 0.91; -- Warlord's of Draenor, 6.1 skill priority adjustments.
-- HoJ.ADDON_VERSION = 0.92; -- Warlord's of Draenor, 6.2 priorities from Summonstone
HoJ.MAJOR_VERSION = 0 -- Battle for Azeroth, 8.1 Tides of Vengeance
-- HoJ.MINOR_VERSION = 93;
-- HoJ.BUILD_VERSION = 1; -- fixed one of the cooldown button timer texts not being set
-- HoJ.BUILD_VERSION = 2; -- fixed a bug where the crusader strike icon was incorrect if the add-on loaded while in prot spec.
-- HoJ.BUILD_VERSION = 3; -- resetting spell icons was resetting empyrean power to not show as divine storm
-- HoJ.BUILD_VERSION = 4; -- removed debug outputs, first beta
-- HoJ.BUILD_VERSION = 5; -- hammer of wrath fixes, added Shield of Vengeance
-- HoJ.BUILD_VERSION = 6; -- selfless healer cooldown
-- HoJ.BUILD_VERSION = 7; -- BUG: sometimes a rotation skill is shown when everything is on cooldown
-- HoJ.BUILD_TYPE = "beta";
-- HoJ.MINOR_VERSION = 95
HoJ.MINOR_VERSION = 96 -- warrior support
--HoJ.BUILD_VERSION = 0; -- Very basic Protection rotation
-- HoJ.BUILD_VERSION = 1; -- Very basic Protection mitigation cooldowns
-- HoJ.BUILD_VERSION = 2; -- Amortized update logic to smooth out cost per frame.
-- HoJ.BUILD_VERSION = 3; -- renamed some frames, added cooldowns
-- HoJ.BUILD_VERSION = 4; -- added charge counter to rotation cooldown, saved variable versioning
-- HoJ.BUILD_VERSION = 5; -- fixed "predictAhead == nil!" spam during ret mode. That's why this is still in beta...
-- HoJ.BUILD_VERSION = 6; -- added data-driven rotation selection for Retribution
-- HoJ.BUILD_VERSION = 7; -- optimizations
-- HoJ.BUILD_VERSION = 8; -- fixed buff timer caching / inquisition never going away if talented.
-- HoJ.BUILD_VERSION = 9; -- Avenger's Shield will show up as a interrupt skill on the cooldowns list if Rebuke is on CD or out of range
-- HoJ.BUILD_VERSION = 10; -- Hammer of Justice should now show up correctly on cooldowns bar when Fist of Justice is talented
-- HoJ.BUILD_VERSION = 11; -- Fixed 10 yard range being calculated using Fist of Justice when FoJ talent was selected
-- HoJ.BUILD_VERSION = 1 -- 0.95.1 - Added frame alpha configuration to addon menu
-- HoJ.BUILD_VERSION = 2 -- 0.95.2 - tracking of blessing buffs on party members and (as much as possible) on non-party members
-- HoJ.BUILD_VERSION = 3 -- 0.95.3 - cleaned up an accidental nested loop in buff check, fixed issue with max party member being out of range
-- HoJ.BUILD_VERSION = 4 -- 0.95.4 - cleaned up range checks, force empyrean power saved var showOnBar to be true to fix invalid saved state
-- HoJ.BUILD_VERSION = 5 -- 0.95.5 - addon versioning via addon chat channels
-- HoJ.BUILD_VERSION = 6 -- 0.95.6 - fix skill bars showing things when dead
-- HoJ.BUILD_VERSION = 7 -- 0.95.7 - fix for old saved variables not getting reset for Empyrean Power
-- HoJ.BUILD_VERSION = 0 -- 0.96.0 - Major refactor + warrior support
-- HoJ.BUILD_VERSION = 1 -- 0.96.1 - Seraphim fix
-- HoJ.BUILD_VERSION = 2 -- 0.96.2 - Russion localization
-- HoJ.BUILD_VERSION = 3 -- 0.96.3 - Fixed priorities on Battle Shout
-- HoJ.BUILD_VERSION = 4 -- 0.96.4 - Fixed next skill not showing. Fixed misses not incrementing num attackers.
HoJ.BUILD_VERSION = 5 -- 0.96.5 - Fixed nil access in damage check

-- HoJ.BUILD_TYPE = "Alpha"
HoJ.BUILD_TYPE = "Beta"

HoJ._savedVersionMajor = HoJ.MAJOR_VERSION
HoJ._savedVersionMinor = HoJ.MINOR_VERSION
HoJ._savedVersionBuild = HoJ.BUILD_VERSION

HoJ._addonPrefix = "HOJ"
HoJ._versionBroadcastInterval = 60 * 5 -- after entering world and every five minutes
HoJ._versionBroadcastNextTime = 0.0
HoJ._versionMessageInterval = 60 * 60 -- once version message has been shown, throttle it to once per hour
HoJ._versionMessageShown = false
HoJ._versionMessageNextTime = 0.0
HoJ._versionBroadcastEnabled = true -- true to send version to other clients
HoJ._versionChecksEnabled = true 	-- true to check versions from other clients and message about new versions

HoJ.SPELL_BUFF = 1			-- buffs a unit
HoJ.SPELL_SHAPESHIFT = 2	-- buffs self
HoJ.SPELL_SKILL = 3			-- not a buff, usually part of rotation
HoJ.SPELL_TALENT = 4		-- not a buff, usually part of rotation, but acquired via talents
HoJ.SPELL_TALENT_BUFF = 5	-- a buff that is a spell and talented (selfless healer)
HoJ.SPELL_AOE_DAMAGE = 6    -- a spell doing damage from an AoE spell, like consecration

HoJ.RANGE_NONE = 1
HoJ.RANGE_SELF = 2
HoJ.RANGE_MELEE = 3
HoJ.RANGE_8_YARDS = 4
HoJ.RANGE_10_YARDS = 5
HoJ.RANGE_12_YARDS = 6
HoJ.RANGE_20_YARDS = 7
HoJ.RANGE_25_YARDS = 8
HoJ.RANGE_30_YARDS = 9
HoJ.RANGE_40_YARDS = 10
HoJ._rangeNames = { "none", "self", "melee", "8", "10", "12", "20", "25", "30", "40" }

-- paladin
HoJ.ROLE_HEALING = 1
HoJ.ROLE_TANK = 2
HoJ.ROLE_DPS = 3
HoJ.ROLE_UNKNOWN = 4
HoJ.ROLE_ANY = 5 		-- for talents usable by any role

HoJ.SPEC_PALADIN_RETRIBUTION = 3
HoJ.SPEC_PALADIN_PROTECTION = 2
HoJ.SPEC_PALADIN_HOLY = 1

HoJ.SPEC_WARRIOR_ARMS = 1
HoJ.SPEC_WARRIOR_FURY = 2
HoJ.SPEC_WARRIOR_PROTECTION = 3

HoJ.CLASS_ID_NONE = 0
HoJ.CLASS_ID_PALADIN = 1
HoJ.CLASS_ID_WARRIOR = 2
HoJ.CLASS_ID_OTHER = 999

HoJ.CLASS_NAME_NONE = "NONE"
HoJ.CLASS_NAME_PALADIN = "PALADIN"
HoJ.CLASS_NAME_WARRIOR = "WARRIOR"

HoJ.SPEC_NAME_WARRIOR_ARMS = "ARMS"
HoJ.SPEC_NAME_WARRIOR_FURY = "FURY"
HoJ.SPEC_NAME_WARRIOR_PROTECTION = "PROTECTION"

HoJ.SPEC_NAME_PALADIN_HOLY = "HOLY"
HoJ.SPEC_NAME_PALADIN_PROTECTION = "PROTECTION"
HoJ.SPEC_NAME_PALADIN_RETRIBUTION = "RETRIBUTION"

-- maps a Blizzard class name to a HoJ class id
HoJ._classNameToClassId = {
    [HoJ.CLASS_NAME_PALADIN] = HoJ.CLASS_ID_PALADIN,
    [HoJ.CLASS_NAME_WARRIOR] = HoJ.CLASS_ID_WARRIOR
}

-- maps a HoJ class id to a Blizzard class name
HoJ._classIdToClassName = {
    [HoJ.CLASS_ID_NONE] = HoJ.CLASS_NAME_NONE,
    [HoJ.CLASS_ID_PALADIN] = HoJ.CLASS_NAME_PALADIN,
    [HoJ.CLASS_ID_WARRIOR] = HoJ.CLASS_NAME_WARRIOR
}

-- maps a HoJ spec name to a Blizzard spec id from GetSpecialization()
HoJ._specNameToSpecId = {
    [HoJ.SPEC_NAME_PALADIN_HOLY] = HoJ.SPEC_PALADIN_HOLY,
    [HoJ.SPEC_NAME_PALADIN_PROTECTION] = HoJ.SPEC_PALADIN_PROTECTION,
    [HoJ.SPEC_NAME_PALADIN_RETRIBUTION] = HoJ.SPEC_PALADIN_RETRIBUTION,
    [HoJ.SPEC_NAME_WARRIOR_ARMS] = HoJ.SPEC_WARRIOR_ARMS,
    [HoJ.SPEC_NAME_WARRIOR_FURY] = HoJ.SPEC_WARRIOR_FURY,
    [HoJ.SPEC_NAME_WARRIOR_PROTECTION] = HoJ.SPEC_WARRIOR_PROTECTION,
}

-- maps a blizzard spec id from GetSpecialization() for a class to a spec name
-- spec names are used to look up entries for a class spec in the HoJ table.
-- the name is used instead of the spec id so that it can also be used to save
-- to the savedVariables and still be human-readable.
HoJ._specIdToSpecName = {
    [HoJ.CLASS_NAME_PALADIN] = {
        [HoJ.SPEC_PALADIN_HOLY] = HoJ.SPEC_NAME_PALADIN_HOLY,
        [HoJ.SPEC_PALADIN_PROTECTION] = HoJ.SPEC_NAME_PALADIN_PROTECTION,
        [HoJ.SPEC_PALADIN_RETRIBUTION] = HoJ.SPEC_NAME_PALADIN_RETRIBUTION,
    },
    [HoJ.CLASS_NAME_WARRIOR] = {
        [HoJ.SPEC_WARRIOR_ARMS] = HoJ.SPEC_NAME_WARRIOR_ARMS,
        [HoJ.SPEC_WARRIOR_FURY] = HoJ.SPEC_NAME_WARRIOR_FURY,
        [HoJ.SPEC_WARRIOR_PROTECTION] = HoJ.SPEC_NAME_WARRIOR_PROTECTION,
    }
}

HoJ._classFunctionTable = {
    [HoJ.CLASS_NAME_PALADIN] = {
        _loadFn = HoJ.LoadPaladin
    },
    [HoJ.CLASS_NAME_WARRIOR] = {
        _loadFn = HoJ.LoadWarrior
    },
}

HoJ._classId = HoJ.CLASS_NONE
HoJ._className = HoJ.CLASS_NAME_NONE
HoJ._defaultSpecId = nil -- based on class is set to a default spec id (pre level 10)
HoJ._classSpecInfo = nil -- reference to table about a class's specs
HoJ._curSpecName = ""
HoJ._curSpecInfo = nil -- reference to the _classSpecInfo sub-table for the current spec

-- a minor heal ratio is the ratio of current/max health where a minor heal will
-- show in a rotation.
-- a major heal ratio is the ratio of current/max health where a major heal will
-- show in a rotation.
-- These are global tables for save / load variables and are synced to the
-- curSpecInfo._healRatio on variable load and when a slider is changed.
HoJ.HEAL_NAME_MINOR = "_minorHealHealthRatio"
HoJ.HEAL_NAME_MAJOR = "_majorHealHealthRatio"

--[[
The world state is used to prevents weird stuff from happening between the
time when the player leaves the world and returns (for instance, taking a portal).
I have seen the following happen:
- logged in on a character level 117
- used party sync and leveled to 120
- never logged out
- when taking a portal, on PLAYER_LEAVING_WORLD, UnitLevel( "player" ) returns 117
- after entering the world again, on PLAYER_ENTERING_WORLD, UnitLevel( "player" ) returns 120
- This was messing with blessing initialization.
--]]

HoJ.WORLD_STATE_UNKNOWN = 0
HoJ.WORLD_STATE_ENTERED = 1
HoJ.WORLD_STATE_LEFT = 2
HoJ._worldState = HoJ.WORLD_STATE_UNKNOWN

HoJ.REACTION_THRESHOLD = 5 -- 4 will only show skill bar for red mobs, 5 will show skill bar for yellow mobs

HoJ.SKILL_USE_PREDICTION_TIME = 0.01 -- look this far beyond the remaining GCD to predict a skill if we don't have one now

HoJ.ROLE_NAMES = { "HEALING", "TANK", "DPS", "UNKNOWN" }
HoJ.STR_TAB = "	"
HoJ.STR_WHITE = "|cffffffff"
HoJ.STR_YELLOW = "|cffffff00"
HoJ.STR_GREEN = "|cff00ff00"
HoJ.STR_PURPLE = "|cffff00ff"
HoJ.STR_RED = "|cffff0000"
HoJ.STR_ORANGE = "|cffff7f00"
HoJ.STR_LIGHTGRAY = "|cffcfcfcf"
HoJ.STR_GRAY = "|cff7f7f7f"
HoJ.STR_WHISPER = "|cffff7fff"
HoJ.STR_CHANNEL = "|cffffc0c0"

HoJ._initialized = false -- false until HoJ:Initialize() is called the first time
HoJ._finalized = false -- fails until HoJ:Finalize() is called

HoJ._needsReset = true	-- resets occur on role switches

HoJ.GROUP_NONE = 0
HoJ.GROUP_PARTY = 1
HoJ.GROUP_RAID = 2
HoJ._groupType = HoJ.GROUP_NONE
HoJ._numGroupMembers = 1

HoJ._role = HoJ.ROLE_UNKNOWN

HoJ.ADDON_NAME = "" -- HoJ.STR_YELLOW .. "Hammer of Justas" .. HoJ.STR_WHITE
HoJ.MSG_PREFIX = "" -- HoJ.STR_YELLOW .. "Hammer of Justas: " .. HoJ.STR_WHITE
HoJ.UPDATES_PER_SEC = 15
HoJ.RENDERS_PER_SEC = 60
HoJ.UPDATE_INTERVAL = 1.0 / HoJ.UPDATES_PER_SEC
HoJ.RENDER_INTERVAL = 1.0 / HoJ.RENDERS_PER_SEC

-- blessing update amortization
HoJ._partyUpdateIndex = -1 	-- amortizes party buff checks by cycling through party members
HoJ._nextUpdateBlessingFrame = 0
HoJ._updateBlessingsInterval = 1 --HoJ.UPDATES_PER_SEC * 5 -- after finishing a party update, wait this number of extra frames before starting the next

-- don't update everything on the same frame
HoJ.UPDATE_START = 1
HoJ.UPDATE_ROTATION = HoJ.UPDATE_START
HoJ.UPDATE_COOLDOWNS = 2
HoJ.UPDATE_BUFFS = 3
HoJ.UPDATE_MAX = HoJ.UPDATE_BUFFS
HoJ._updateType = HoJ.UPDATE_START

HoJ._nextUpdateTime = 0.0
HoJ._nextRenderTime = 0.0

HoJ._isActive = false
HoJ._inCombat = false
HoJ._combatStartTime = 0.0
HoJ._level = 0
HoJ._power = 0
HoJ._lastPower = 0  -- last player power from UNIT_POWER_UPDATE
HoJ._lastPowerGainTime = 0.0 -- last time power was gained
HoJ._nextPowerGainTime = 0.0 -- predicted next power gain time
HoJ._powerGainInterval = 3.0 -- interval between power gain
HoJ._powerGainStep = 25.0   -- amount of power gain per interval
HoJ._powerPerSec = 0
HoJ._totalPowerGain = 0 -- power gained during the current combat session
HoJ.MAX_HOLY_POWER = 5
HoJ._useHolyPower = false 	-- true when a holy power skill should be used
HoJ._playerName = nil
HoJ._playerRaceId = 0
HoJ._playerGUID = nil

HoJ._playerHealth = 1
HoJ._playerMaxHealth = 1
HoJ._playerHealthRatio = 1.0
HoJ._predictedHealth = 1
HoJ._predictedHealthTime = 0.0
HoJ._predictedDeathTime = 999.0 -- 999.0 means no death predicted 
HoJ._predictedNearDeathTime = 999.0 -- 999.0 means no death predicted

HoJ._incomingDps = 0 -- damage per second taken by the player over the last several seconds
HoJ._healsPerSec = 0
HoJ._incomingDpsRatio = 0.0
HoJ._healsPerSecRatio = 0.0
HoJ._largestRecentDamageRatio = 0.0

HoJ._nextHealthTime = 0.0
HoJ._showHealthTime = 0.0	-- set to a time in the future to see health spam

HoJ.HEALTH_SAMPLE_SECONDS = 3 -- number of seconds we keep health history for
HoJ.HEALTH_SAMPLES_PER_SEC = 15	-- number of times per second that we sample health (make sure it's less than HoJ.UPDATE_INTERVAL)

HoJ._frameNum = 0 -- updated for every call to HoJ_OnUpdate
HoJ._updateFrameNum = 0 -- updated for every call to HoJ:Update
HoJ._renderFrameNum = 0 -- updated for every call to HoJ:UpdateRender
HoJ._enteredWorldUpdateFrameNum = 0 -- frame when player entered world, used to gate some udpates after entering world until some things settle
HoJ._updateElapsedTime = 0.0
HoJ._renderElapsedTime = 0.0
HoJ._updateBuffsTime = 0.0 -- when to next check buffs
HoJ._updateCooldownsTime = 0.0 -- when to next check cooldowns
--HoJ._divinePleaTriggerPercentage = 0.75;	-- trigger when we have less than this % of max mana
HoJ._blockedNames = { }	--["testname"] = "testname" };
--HoJ._blockedNames = { ["Mienyu"] = "Mienyu", ["Justas"] = "Justas" };
HoJ._ignoreHistory = { }
HoJ._silentIgnore = false
HoJ._stealthIgnore = true
HoJ._stealthIgnoreMessage = HoJ.STR_CHANNEL .. "[3. WorldDefense] : " .. HoJ.STR_YELLOW .. "Isle of Quel'Danas is under attack!"
HoJ._maxGlowAlpha = 0.5
HoJ._minGlowAlpha = 0.1
HoJ._glowAlpha = HoJ._minGlowAlpha
HoJ._glowScale = 1.025
HoJ._skillCDAlpha = 0.25;	-- alpha for spell icons on cooldown
HoJ._skillNormalAlpha = 0.75	-- alpha for skill icons not on cooldown
HoJ._events = { }
HoJ._HoPo = { nil, nil, nil, nil, nil }
HoJ._HoPoGlow = { nil, nil, nil, nil, nil }
HoJ._curTime = GetTime()
HoJ._lookAheadTime = 0.0 -- sometimes we want to perform a test at some future time. This is the extra time to add to _curTime.
HoJ._gcdRemaining = 0.0
HoJ._gcdDuration = 0.0 -- we cache this because we can't just query it when a spell isn't active

-- combat dps persist during active combat
HoJ._combatDpsStats = {
    _startTime = 0.0,
    _totalTime = nil, -- unused for combat dps
    _damageDone = 0,
    _attacksHit = 0,
    _attacksHitThreshold = 5,
    _timeThreshold = 5.0,
    _dps = 0.0,
    _maxDps = 0.0,
    _isValid = false,
    _targetTimeToLive = -1.0,
    _firstPredictedTargetDeathTime = -1.0,
    _predictedTargetDeathTime = -1.0,
}

HoJ._sessionDpsStats = {
    _startTime = nil, -- unused for session dps
    _totalTime = 0.0,
    _damageDone = 0,
    _attacksHit = 0,
    _attacksHitThreshold = 5,
    _timeThreshold = 5.0,
    _dps = 0.0,
    _maxDps = 0.0,
    _isValid = false,
    _targetTimeToLive = -1.0,
    _firstPredictedTargetDeathTime = -1.0,
    _predictedTargetDeathTime = -1.0,
}

HoJ._targetInfo = {
    _name = ""
}

HoJ._targetCanAttack = false
HoJ._targetCanBeInterrupted = false
HoJ._targetHealth = 0.0
HoJ._targetMaxHealth = 0.0
HoJ._targetHealthRatio = 0.0
HoJ._targetLevel = 1
HoJ._targetClassification = ""
HoJ._targetIsTough = false
HoJ._targetIsBoss = false -- "worldboss" or "rareelite"

HoJ._targetInRange = { }	-- one entry for each HoJ.RANGE_ type, indicating if the target is within that range

HoJ._showSkillCDs = true
HoJ._showCooldownsFrame = true
HoJ._showBuffsFrame = true
HoJ._showTemplarsVerdictFrame = true
HoJ._showHolyPowerFrame = true
HoJ._showRotationFrame = true
HoJ._showRotationSecondaryFrame = true
HoJ._showDebugFrame = false

HoJ._enabled = true;

HoJ.INCOMING_DPS_CHECK_INTERVAL = 2.0
HoJ.AVENGING_WRATH_INCOMING_DPS_RATIO = -0.10	-- incoming DPS less than this ratio will trigger avenging wrath icon
HoJ.AVENGING_WRATH_HEALTH_THRESHOLD = 0.5
HoJ.PROT_GUARDIAN_INCOMING_DPS_RATIO = -0.2 -- incoming DPS less than this will trigger guardian of ancient kings icon

HoJ._buffs = { }	-- list of currently active buffs
HoJ._buffFrames = { }

HoJ._cooldowns = { }	-- list of skills to execute in order of priority / cooldown
HoJ._cooldownFrames = { }

HoJ._rotationSkills = { } -- list of best skills (only ever 1)
HoJ._rotationFrames = { }

HoJ._windowFrames = { }  -- list of all HoJ-specific window frames
HoJ._omniCC = false -- true if omniCC is enabled (it will show it's own timer)

HoJ._globalAlpha = 1.0 -- global alpha for HoJ UI frames
HoJ._rotationAlpha = 1.0
HoJ._rotationSecondaryAlpha = 0.75
HoJ._cooldownsAlpha = 1.0
HoJ._buffsAlpha = 1.0
HoJ._holyPowerAlpha = 1.0
HoJ._debugAlpha = 1.0 

HoJ._debugSpellName = nil -- set to a spell name, e.g. "Wake of Ashes", to debug that spell
HoJ._showRange = false
HoJ._debugBlessings = false
HoJ._debugRulesLevel = 0
HoJ._debugNextSkillLevel = 0
HoJ._debugPower = false

HoJ._throttledMessages = { }	-- list of throttled messages

HoJ._text = nil

HoJ._recentMeleeAttackers = { }
HoJ._nextRecentMeleeAttackerUpdateFrame = 0
HoJ._numMeleeAttackers = 0
HoJ._lastSpellDamageTime = 0.0

HoJ._controlLost = false
HoJ._hasDebuff = false